/*
	Copyright (C) 2004 Christopher E. Miller
	
	This software is provided 'as-is', without any express or implied
	warranty.  In no event will the authors be held liable for any damages
	arising from the use of this software.
	
	Permission is granted to anyone to use this software for any purpose,
	including commercial applications, and to alter it and redistribute it
	freely, subject to the following restrictions:
	
	1. The origin of this software must not be misrepresented; you must not
	   claim that you wrote the original software. If you use this software
	   in a product, an acknowledgment in the product documentation would be
	   appreciated but is not required.
	2. Altered source versions must be plainly marked as such, and must not be
	   misrepresented as being the original software.
	3. This notice may not be removed or altered from any source distribution.
*/


version(Linux)
{
	version = BsdSockets;
}

version(Win32)
{
	typedef uint socket_t = ~0;
}
else version(BsdSockets)
{
	typedef int socket_t = -1;
}
else
{
	static assert(0); // No socket support yet
}
const socket_t INVALID_SOCKET = socket_t.init;
const int SOCKET_ERROR = -1;


private:

import std.string, std.stdint;


version(Win32)
{
	import std.c.windows.windows;
	
	
	extern(Windows)
	{
		const int WSADESCRIPTION_LEN = 256;
		const int WSASYS_STATUS_LEN = 128;
		
		struct WSADATA
		{
			WORD wVersion;
			WORD wHighVersion;
			char szDescription[WSADESCRIPTION_LEN+1];
			char szSystemStatus[WSASYS_STATUS_LEN+1];
			ushort iMaxSockets;
			ushort iMaxUdpDg;
			char* lpVendorInfo;
		}
		alias WSADATA* LPWSADATA;
		
		
		struct fd_set
		{
			uint fd_count;
			// N socket_t's created on-the-fly
		}
		
		
		const int IOCPARM_MASK =  0x7f;
		const int IOC_IN =        cast(int)0x80000000;
		const int FIONBIO =       IOC_IN | ((uint.size & IOCPARM_MASK) << 16) | (102 << 8) | 126;
		const int SOL_SOCKET =    0xFFFF;
		const int SO_TYPE =       0x1008;
		const uint INADDR_NONE =  0xFFFFFFFF;
		
		
		int WSAStartup(WORD wVersionRequested, LPWSADATA lpWSAData);
		int WSACleanup();
		socket_t socket(int af, int type, int protocol);
		int ioctlsocket(socket_t s, int cmd, uint* argp);
		int getsockopt(socket_t s, int level, int optname, char* optval, int* optlen);
		uint inet_addr(char* cp);
		int bind(socket_t s, sockaddr* name, int namelen);
		int connect(socket_t s, sockaddr* name, int namelen);
		int listen(socket_t s, int backlog);
		socket_t accept(socket_t s, sockaddr* addr, int* addrlen);
		int closesocket(socket_t s);
		int shutdown(socket_t s, int how);
		int getpeername(socket_t s, sockaddr* name, int* namelen);
		int getsockname(socket_t s, sockaddr* name, int* namelen);
		int send(socket_t s, void* buf, int len, int flags);
		int sendto(socket_t s, void* buf, int len, int flags, sockaddr* to, int tolen);
		int recv(socket_t s, void* buf, int len, int flags);
		int recvfrom(socket_t s, void* buf, int len, int flags, sockaddr* from, int* fromlen);
		int select(int nfds, fd_set* readfds, fd_set* writefds, fd_set* errorfds, timeval* timeout);
		//int __WSAFDIsSet(socket_t s, fd_set* fds);
		int getsockopt(socket_t s, int level, int optname, void* optval, int* optlen);
		int setsockopt(socket_t s, int level, int optname, void* optval, int optlen);
		
		
		const int WSAEWOULDBLOCK = 10035;
		int WSAGetLastError();
	}
}
else version(BsdSockets)
{
	extern(C)
	{
		const int F_GETFL =       3;
		const int F_SETFL =       4;
		const int O_NONBLOCK =    0x4000;
		const int SOL_SOCKET =    0xFFFF;
		const int SO_TYPE =       0x1008;
		const uint INADDR_NONE =  cast(uint)-1;
		
		
		socket_t socket(int af, int type, int protocol);
		int fcntl(socket_t s, int f, ...);
		int getsockopt(socket_t s, int level, int optname, char* optval, int* optlen);
		uint inet_addr(char* cp);
		int bind(socket_t s, sockaddr* name, int namelen);
		int connect(socket_t s, sockaddr* name, int namelen);
		int listen(socket_t s, int backlog);
		socket_t accept(socket_t s, sockaddr* addr, int* addrlen);
		int close(socket_t s);
		int shutdown(socket_t s, int how);
		int getpeername(socket_t s, sockaddr* name, int* namelen);
		int getsockname(socket_t s, sockaddr* name, int* namelen);
		int send(socket_t s, void* buf, int len, int flags);
		int sendto(socket_t s, void* buf, int len, int flags, sockaddr* to, int tolen);
		int recv(socket_t s, void* buf, int len, int flags);
		int recvfrom(socket_t s, void* buf, int len, int flags, sockaddr* from, int* fromlen);
		int select(int nfds, fd_set* readfds, fd_set* writefds, fd_set* errorfds, timeval* timeout);
		int getsockopt(socket_t s, int level, int optname, void* optval, int* optlen);
		int setsockopt(socket_t s, int level, int optname, void* optval, int optlen);
		
		
		const int EINPROGRESS = -1; //EWOULDBLOCK
		extern int errno;
	}
}


struct sockaddr
{
	ushort sa_family;               
	char[14] sa_data = [0];             
}


version(BigEndian)
{
	uint16_t htons(uint16_t x)
	{
		return x;
	}
	
	
	uint32_t htonl(uint32_t x)
	{
		return x;
	}
}
else version(LittleEndian)
{
	import std.intrinsic;
	
	
	uint16_t htons(uint16_t x)
	{
		return (x >> 8) | (x << 8);
	}


	uint32_t htonl(uint32_t x)
	{
		return bswap(x);
	}
}
else
{
	// Middle endian:
	// "For developers of new software, it is not only perfectly reasonable, but
	// strongly recommended, to pretend that this doesn't exist." -- Jan Wolter
	static assert(0);
}


uint16_t ntohs(uint16_t x)
{
	return htons(x);
}


uint32_t ntohl(uint32_t x)
{
	return htonl(x);
}


public:
class SocketException: Exception
{
	this(char[] msg)
	{
		super(msg);
	}
}


static this()
{
	version(Win32)
	{
		WSADATA wd;
		if(WSAStartup(0x0101, &wd))
			throw new SocketException("Unable to initialize socket library.");
	}
}


static ~this()
{
	version(Win32)
	{
		WSACleanup();
	}
}


version(Win32)
{
	enum AddressFamily: int
	{
		UNSPEC =     0,
		UNIX =       1,
		INET =       2,
		IPX =        6,
		APPLETALK =  16,
		//INET6 =      ? // Need Windows XP ?
	}
}
else version(BsdSockets)
{
	enum AddressFamily: int
	{
		UNSPEC =     0,
		UNIX =       1,
		INET =       2,
		IPX =        4,
		APPLETALK =  5,
		//INET6 =      10,
	}
}


enum SocketType: int
{
	STREAM =     1,
	DGRAM =      2,
	RAW =        3,
	RDM =        4,
	SEQPACKET =  5,
}


enum ProtocolType: int
{
	IP =    0,
	ICMP =  1,
	IGMP =  2,
	GGP =   3,
	TCP =   6,
	PUP =   12,
	UDP =   17,
	IDP =   22,
	ND =    77,
	RAW =   255,
	MAX =   256,
}


class AddressException: SocketException
{
	this(char[] msg)
	{
		super(msg);
	}
}


interface Address
{
	protected sockaddr* name();
	protected int nameLen();
	AddressFamily addressFamily();
	char[] toString();
}


class UnknownAddress: Address
{
	protected:
	sockaddr sa;
	
	
	sockaddr* name()
	{
		return &sa;
	}
	
	
	int nameLen()
	{
		return sa.size;
	}
	
	
	public:
	AddressFamily addressFamily()
	{
		return cast(AddressFamily)sa.sa_family;
	}
	
	
	char[] toString()
	{
		return "Unknown";
	}
}


class InternetAddress: Address
{
	protected:
	struct sockaddr_in
	{
		ushort sin_family = cast(ushort)AddressFamily.INET;
		ushort sin_port;
		uint sin_addr; //in_addr
		char[8] sin_zero = [0];
	}
	sockaddr_in sin;


	sockaddr* name()
	{
		return cast(sockaddr*)&sin;
	}
	
	
	int nameLen()
	{
		return sin.size;
	}
	
	
	this()
	{
	}
	
	
	public:
	AddressFamily addressFamily()
	{
		return AddressFamily.INET;
	}
	
	
	ushort port()
	{
		return ntohs(sin.sin_port);
	}
	
	
	uint addr()
	{
		return ntohl(sin.sin_addr);
	}
	
	
	//-port- of 0 means any
	//-addr- is an IP address in the format "a.b.c.d"
	this(char[] addr, ushort port)
	{
		uint uiaddr = inet_addr(std.string.toStringz(addr));
		if(uiaddr == INADDR_NONE)
			throw new AddressException("Invalid internet address.");
		sin.sin_addr = uiaddr;
		sin.sin_port = htons(port);
	}
	
	
	const uint ADDR_ANY = 0;
	const uint ADDR_NONE = cast(int)-1;
	const ushort PORT_ANY = 0;
	
	
	this(uint addr, ushort port)
	{
		sin.sin_addr = htonl(addr);
		sin.sin_port = htons(port);
	}
	
	
	this(ushort port)
	{
		sin.sin_addr = 0; //any, "0.0.0.0"
		sin.sin_port = htons(port);
	}
	
	
	char[] toAddrString()
	{
		char[] result = new char[16];
		int i = 0;
		char* sip = (char*)&sin.sin_addr;
		char[] s;

		s = std.string.toString((uint)sip[0]);
		result[i .. i + s.length] = s;
		i += s.length;
		result[i] = '.';
		i++;

		s = std.string.toString((uint)sip[1]);
		result[i .. i + s.length] = s;
		i += s.length;
		result[i] = '.';
		i++;

		s = std.string.toString((uint)sip[2]);
		result[i .. i + s.length] = s;
		i += s.length;
		result[i] = '.';
		i++;

		s = std.string.toString((uint)sip[3]);
		result[i .. i + s.length] = s;
		i += s.length;

		return result[0 .. i];
	}
	
	
	char[] toPortString()
	{
		return std.string.toString(port());
	}
	
	
	char[] toString()
	{
		return toAddrString() ~ ":" ~ toPortString();
	}
}


unittest
{
	InternetAddress ia = new InternetAddress("63.105.9.61", 80);
	assert(ia.toString() == "63.105.9.61:80");
}


class SocketAcceptException: SocketException
{
	this(char[] msg)
	{
		super(msg);
	}
}


enum SocketShutdown: int
{
	RECEIVE =  0,
	SEND =     1,
	BOTH =     2,
}


enum SocketFlags: int
{
	NONE =           0,
	OOB =            0x1, //out of band
	PEEK =       0x02, //only for receiving
	DONTROUTE =  0x04, //only for sending
}


struct timeval
{
	int tv_sec; //seconds
	int tv_usec; //microseconds
}


//a set of sockets for Socket.select()
class SocketSet
{
	private:
	byte* fds; //includes index 0 count
	uint reserved;
	
	
	//setter
	void count(uint n)
	{
		*(cast(uint*)&fds[0]) = n;
	}
	
	
	socket_t* first()
	{
		return cast(socket_t*)(fds + uint.size);
	}
	
	
	public:
	//must be called if reusing
	void reset()
	{
		//reset(64); //winsock standard
		reset(8); //since it's growable
	}
	
	
	void reset(uint reserve)
	{
		if(!reserve)
			reserve = 1;
		fds = new byte[reserve * socket_t.size + uint.size];
		count = 0;
		reserved = reserve;
	}
	
	
	void add(Socket s)
	{
		add(s.sock);
	}
	
	
	void add(socket_t s)
	in
	{
		assert(count <= reserved);
	}
	body
	{
		uint c = count;
		if(c == reserved)
		{
			int nr = reserved * 2;
			byte* nfds = new byte[nr * socket_t.size + uint.size];
			
			nfds[0 .. reserved] = fds[0 .. reserved];
			fds = nfds;
			reserved = nr;
		}
		first()[c] = s;
		count = c + 1;
	}
	
	
	void remove(Socket s)
	{
		remove(s.sock);
	}
	
	
	void remove(socket_t s)
	{
		socket_t* start, stop;
		start = first();
		stop = start + count;
		for(; start != stop; start++)
		{
			if(*start == s)
				goto found;
		}
		return;
		found:
		for(++start; start != stop; start++)
		{
			*(start - 1) = *start;
		}
		count = count - 1; // count--; gives error: 'this.count' is not a scalar, it is a void(uint n)
	}
	
	
	//getter
	uint count()
	{
		return *(cast(uint*)&fds[0]);
	}
	
	
	fd_set* toFd_set()
	{
		return cast(fd_set*)fds;
	}
	
	
	bit isSet(socket_t s)
	{
		socket_t* start, stop;
		start = first();
		stop = start + count;
		for(; start != stop; start++)
		{
			if(*start == s)
				return true;
		}
		return false;
	}
	
	
	bit isSet(Socket s)
	{
		return isSet(s.sock);
	}
	
	
	//one could foreach() socket_t's and lookup Socket's
	//in an associative array instead of calling isSet
	//but it seems like a hack...
	int opApply(int delegate(inout socket_t) dg)
	{
		int result = 0;
		socket_t* start, stop;
		start = first();
		stop = start + count;
		for(; start != stop; start++)
		{
			result = dg(*start);
			if(result)
				break;
		}
		return result;
	}
	
	
	this()
	{
		reset();
	}
	
	
	//initial reserve
	this(uint reserve)
	{
		reset(reserve);
	}
	
	
	SocketSet dup()
	{
		SocketSet ss = new SocketSet;
		ss.fds = fds[0 .. reserved * socket_t.size + uint.size].dup;
		return ss;
	}
}


enum SocketOptionLevel
{
	SOCKET =  0xFFFF, //different source 1
	IP =      0,
	TCP =     6,
	UDP =     17,
}


struct linger
{
	ushort l_onoff; //option on/off
	ushort l_linger; //linger time
}


enum SocketOption: int
{
	//consistent
	SO_DEBUG =         0x1,
	
	//possibly Winsock-only values
		SO_BROADCAST =  0x20,
		SO_REUSEADDR =  0x4,
		SO_LINGER =     0x80,
		SO_OOBINLINE =  0x100,
		SO_SNDBUF =     0x1001,
		SO_RCVBUF =     0x1002,
		SO_ERROR =      0x1007,
	
	/*
	//different source
		SO_BROADCAST =  -1,
		SO_REUSEADDR =  0x406,
		SO_LINGER =     -3,
		SO_OOBINLINE =  -4,
		SO_SNDBUF =     3,
		SO_RCVBUF =     2,
		SO_ERROR =      9,
	*/
	
	SO_ACCEPTCONN =    0x2, // ?
	SO_KEEPALIVE =     0x8, // ?
	SO_DONTROUTE =     0x10, // ?
	SO_TYPE =          0x1008, // ?
}


class Socket
{
	private:
	socket_t sock;
	AddressFamily _family;
	
	version(Win32)
		bit _blocking = false;
	
	
	this(socket_t sock)
	{
		this.sock = sock;
	}
	
	
	public:
	this(AddressFamily af, SocketType type, ProtocolType protocol)
	{
		sock = socket(af, type, protocol);
		if(sock == sock.init)
			throw new SocketException("Unable to create socket.");
		_family = af;
	}
	
	
	~this()
	{
		close();
	}
	
	
	//get underlying socket handle
	socket_t handle()
	{
		return sock;
	}
	
	
	override char[] toString()
	{
		return "Socket";
	}
	
	
	//getter
	bit blocking()
	{
		version(Win32)
		{
			return _blocking;
		}
		else version(BsdSockets)
		{
			return !(fcntl(handle, F_GETFL, 0) & O_NONBLOCK);
		}
	}
	
	
	//setter
	void blocking(bit byes)
	{
		version(Win32)
		{
			uint num = !byes;
			if(SOCKET_ERROR == ioctlsocket(sock, FIONBIO, &num))
				goto err;
			_blocking = byes;
		}
		else version(BsdSockets)
		{
			int x = fcntl(handle, F_GETFL, 0);
			if(byes)
				x &= ~O_NONBLOCK;
			else
				x |= O_NONBLOCK;
			if(SOCKET_ERROR == fcntl(sock, F_SETFL, x))
				goto err;
		}
		return; //success
		
		err:
		throw new SocketException("Unable to set socket blocking.");
	}
	
	
	AddressFamily addressFamily()
	{
		return _family;
	}
	
	
	bit isAlive()
	{
		int type, typesize = type.size;
		return !getsockopt(sock, SOL_SOCKET, SO_TYPE, (char*)type, &typesize);
	}
	
	
	void bind(Address addr)
	{
		if(SOCKET_ERROR == .bind(sock, addr.name(), addr.nameLen()))
			throw new SocketException("Unable to bind socket.");
	}
	
	
	void connect(Address to)
	{
		if(SOCKET_ERROR == .connect(sock, to.name(), to.nameLen()))
		{
			if(!_blocking)
			{
				version(Win32)
				{
					if(WSAEWOULDBLOCK == WSAGetLastError())
						return;
				}
				else version(BsdSockets)
				{
					if(EINPROGRESS == errno)
						return;
				}
			}
			throw new SocketException("Unable to connect socket.");
		}
	}
	
	
	//need to bind() first
	void listen(int backlog)
	{
		if(SOCKET_ERROR == .listen(sock, backlog))
			throw new SocketException("Unable to listen on socket.");
	}
	
	
	Socket accept()
	{
		socket_t newsock = .accept(sock, null, null);
		if(INVALID_SOCKET == newsock)
			throw new SocketAcceptException("Unable to accept socket connection.");
		Socket newSocket = new Socket(newsock);
		version(Win32)
			newSocket._blocking = _blocking; //inherits blocking mode
		newSocket._family = _family; //same family
		return newSocket;
	}
	
	
	void shutdown(SocketShutdown how)
	{
		.shutdown(sock, cast(int)how);
	}
	
	
	//calling shutdown() before this is recommended
	//for connection-oriented sockets
	void close()
	{
		version(Win32)
		{
			.closesocket(sock);
		}
		else version(BsdSockets)
		{
			.close(sock);
		}
	}
	
	
	private Address newFamilyObject()
	{
		Address result;
		switch(_family)
		{
			case AddressFamily.INET:
				result = new InternetAddress;
				break;
			
			default:
				result = new UnknownAddress;
		}
		return result;
	}
	
	
	Address remoteAddress()
	{
		Address addr = newFamilyObject();
		int nameLen = addr.nameLen();
		if(SOCKET_ERROR == .getpeername(sock, addr.name(), &nameLen))
			throw new SocketException("Unable to obtain remote socket address.");
		assert(addr.addressFamily() == _family);
		return addr;
	}
	
	
	Address localAddress()
	{
		Address addr = newFamilyObject();
		int nameLen = addr.nameLen();
		if(SOCKET_ERROR == .getsockname(sock, addr.name(), &nameLen))
			throw new SocketException("Unable to obtain local socket address.");
		assert(addr.addressFamily() == _family);
		return addr;
	}
	
	
	private void cannotSend(int count)
	{
		throw new SocketException("Unable to send " ~ std.string.toString(count) ~ " bytes.");
	}
	
	
	//returns number of bytes actually sent
	int send(void[] buf, SocketFlags flags)
	{
		int sent = .send(sock, buf, buf.length, cast(int)flags);
		if(SOCKET_ERROR == sent)
			cannotSend(buf.length);
		return sent;
	}
	
	
	int send(void[] buf)
	{
		return send(buf, SocketFlags.NONE);
	}
	
	
	//-to- is ignored if connected ?
	int sendTo(void[] buf, SocketFlags flags, Address to)
	{
		int sent = .sendto(sock, buf, buf.length, cast(int)flags, to.name(), to.nameLen());
		if(SOCKET_ERROR == sent)
			cannotSend(buf.length);
		return sent;
	}
	
	
	//-to- is ignored if connected ?
	int sendTo(void[] buf, Address to)
	{
		return sendTo(buf, SocketFlags.NONE, to);
	}
	
	
	//assumes you connect()ed
	int sendTo(void[] buf, SocketFlags flags)
	{
		int sent = .sendto(sock, buf, buf.length, cast(int)flags, null, 0);
		if(SOCKET_ERROR == sent)
			cannotSend(buf.length);
		return sent;
	}
	
	
	//assumes you connect()ed
	int sendTo(void[] buf)
	{
		return sendTo(buf, SocketFlags.NONE);
	}
	
	
	private void cannotReceive()
	{
			throw new SocketException("Unable to receive socket data.");
	}
	
	
	//returns number of bytes actually received
	int receive(void[] buf, SocketFlags flags)
	{
		if(!buf.length) //return 0 and don't think the connection closed
			return 0;
		int read = .recv(sock, buf, buf.length, cast(int)flags);
		if(read == -1)
			cannotReceive();
		// if(!read) //connection closed
		return read;
	}
	
	
	int receive(void[] buf)
	{
		return receive(buf, SocketFlags.NONE);
	}
	
	
	//-from- is ignored if connected ?
	int receiveFrom(void[] buf, SocketFlags flags, out Address from)
	{
		if(!buf.length) //return 0 and don't think the connection closed
			return 0;
		from = newFamilyObject();
		int nameLen = from.nameLen();
		int read = .recvfrom(sock, buf, buf.length, cast(int)flags, from.name(), &nameLen);
		if(read == -1)
			cannotReceive();
		assert(from.addressFamily() == _family);
		// if(!read) //connection closed
		return read;
	}
	
	
	//-from- is ignored if connected ?
	int receiveFrom(void[] buf, out Address from)
	{
		return receiveFrom(buf, SocketFlags.NONE, from);
	}
	
	
	//assumes you connect()ed
	int receiveFrom(void[] buf, SocketFlags flags)
	{
		if(!buf.length) //return 0 and don't think the connection closed
			return 0;
		int read = .recvfrom(sock, buf, buf.length, cast(int)flags, null, null);
		if(read == -1)
			cannotReceive();
		// if(!read) //connection closed
		return read;
	}
	
	
	//assumes you connect()ed
	int receiveFrom(void[] buf)
	{
		return receiveFrom(buf, SocketFlags.NONE);
	}
	
	
	//returns the length, in bytes, of the actual result - very different from getsockopt()
	int getOption(SocketOptionLevel level, SocketOption option, void[] result)
	{
		int len = result.length;
		if(SOCKET_ERROR == .getsockopt(sock, cast(int)level, cast(int)option, result, &len))
			throw new SocketException("Unable to get socket option.");
		return len;
	}
	
	
	void setOption(SocketOptionLevel level, SocketOption option, void[] value)
	{
		if(SOCKET_ERROR == .setsockopt(sock, cast(int)level, cast(int)option, value, value.length))
			throw new SocketException("Unable to set socket option.");
	}
	
	
	//SocketSet's updated to include only those sockets which an event occured
	//returns the number of events, or 0 on timeout
	//for a connect()ing socket, writeability means connected
	//for a listen()ing socket, readability means listening
	//Winsock: possibly internally limited to 64 sockets per set
	static int select(SocketSet checkRead, SocketSet checkWrite, SocketSet checkError, timeval* tv)
	in
	{
		//make sure none of the SocketSet's are the same object
		if(checkRead)
		{
			assert(checkRead !== checkWrite);
			assert(checkRead !== checkError);
		}
		if(checkWrite)
		{
			assert(checkWrite !== checkError);
		}
	}
	body
	{
		int result = .select(int.max, checkRead ? checkRead.toFd_set() : null,
			checkWrite ? checkWrite.toFd_set() : null,
			checkError ? checkError.toFd_set() : null,
			tv);
		if(SOCKET_ERROR == result)
			throw new SocketException("Socket select error.");
		return result;
	}
	
	
	static int select(SocketSet checkRead, SocketSet checkWrite, SocketSet checkError, int microseconds)
	{
		timeval tv;
		tv.tv_sec = 0;
		tv.tv_usec = microseconds;
		return select(checkRead, checkWrite, checkError, &tv);
	}
	
	
	//maximum timeout
	static int select(SocketSet checkRead, SocketSet checkWrite, SocketSet checkError)
	{
		return select(checkRead, checkWrite, checkError, null);
	}
	
	
	/+
	bit poll(events)
	{
		int WSAEventSelect(socket_t s, WSAEVENT hEventObject, int lNetworkEvents); // Winsock 2 ?
		int poll(pollfd* fds, int nfds, int timeout); // Unix ?
	}
	+/
}

